#!/usr/bin/env python

"""Tests for `revel2d2l` package."""


import unittest

# from revel2d2l import revel2d2l


class TestRevel2d2l(unittest.TestCase):
    """Tests for `revel2d2l` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""
